"""Data structures, constants and enumerations needed to interact with a Myo armband."""
from __future__ import annotations

__all__ = [
    "Pose",
    "SKU",
    "HardwareRev",
    "FirmwareVersion",
    "EmgMode",
    "EmgValue",
    "ImuMode",
    "ClassifierMode",
    "VibrationType",
    "VIBRATE2_STEPS",
    "SleepMode",
    "UnlockType",
    "UserActionType",
    "ClassifierModelType",
    "MotionEventType",
    "ClassifierEventType",
    "Arm",
    "XDirection",
    "SyncResult",
    "FirmwareInfo",
    "EMGCallback",
    "EMGSmoothCallback",
    "Quaternion",
    "IMUCallback",
    "LockCallback",
    "PoseCallback",
    "SyncCallback",
    "TapCallback",
]

from enum import IntEnum
from typing import TYPE_CHECKING, Final, NamedTuple, Protocol, Tuple

if TYPE_CHECKING:
    import sys

    if sys.version_info < (3, 10):
        from typing_extensions import TypeAlias
    else:
        from typing import TypeAlias


class Pose(IntEnum):
    """Supported poses."""

    REST = 0x00
    FIST = 0x01
    WAVE_IN = 0x02
    WAVE_OUT = 0x03
    FINGERS_SPREAD = 0x04
    DOUBLE_TAP = 0x05
    UNKNOWN = 0xFF


class SKU(IntEnum):
    """Known Myo SKUs.

    Attributes:
        UNKNOWN: Unknown SKU (default value for old firmwares).
        BLACK_MYO: Black Myo.
        WHITE_MYO: White Myo.
    """

    UNKNOWN = 0
    BLACK_MYO = 1
    WHITE_MYO = 2


class HardwareRev(IntEnum):
    """Known Myo hardware revisions.

    Attributes:
        UNKNOWN: Unknown hardware revision.
        REV_C: Myo Alpha (REV-C) hardware.
        REV_D: Myo (REV-D) hardware.
    """

    UNKNOWN = 0
    REV_C = 1
    REV_D = 2


class FirmwareVersion(NamedTuple):
    """Version information for the Myo firmware.

    Attributes:
        major: Major version.
        minor: Minor version. It is incremented for changes in this interface.
        patch: Patch version. It is incremented for firmware changes that do not
            introduce changes in this interface.
        hardware_rev: Myo hardware revision.
    """

    major: int
    minor: int
    patch: int
    hardware_rev: HardwareRev


class EmgMode(IntEnum):
    """EMG modes.

    Attributes:
        NONE: Do not send EMG data.
        SMOOTH: Undocumented mode. Send 50 Hz rectified and smoothed positive values
            correlated with muscle 'activation'.
        EMG: Send filtered EMG data.
        EMG_RAW: Send raw (unfiltered) EMG data.
    """

    NONE = 0x00
    SMOOTH = 0x01
    EMG = 0x02
    EMG_RAW = 0x03


class ImuMode(IntEnum):
    """IMU modes.

    Attributes:
        NONE: Do not send IMU data or events.
        DATA: Send IMU data streams (accelerometer, gyroscope, and orientation).
        EVENTS: Send motion events detected by the IMU (e.g. taps).
        ALL: Send both IMU data streams and motion events.
        RAW: Send raw IMU data streams.
    """

    NONE = 0x00
    DATA = 0x01
    EVENTS = 0x02
    ALL = 0x03
    RAW = 0x04


class ClassifierMode(IntEnum):
    """Classifier modes.

    Attributes:
        DISABLED: Disable and reset the internal state of the onboard classifier.
        ENABLED: Send classifier events (poses and arm events).
    """

    DISABLED = 0x00
    ENABLED = 0x01


class VibrationType(IntEnum):
    """Kinds of vibrations.

    Attributes:
        NONE: Do not vibrate.
        SHORT: Vibrate for a short amount of time.
        MEDIUM: Vibrate for a medium amount of time.
        LONG: Vibrate for a long amount of time.
    """

    NONE = 0x00
    SHORT = 0x01
    MEDIUM = 0x02
    LONG = 0x03


VIBRATE2_STEPS: Final = 6


class SleepMode(IntEnum):
    """Sleep modes.

    Attributes:
        NORMAL: Normal sleep mode; Myo will sleep after a period of inactivity.
        NEVER_SLEEP: Never go to sleep.
    """

    NORMAL = 0
    NEVER_SLEEP = 1


class UnlockType(IntEnum):
    """Unlock types.

    Attributes:
        LOCK: Re-lock immediately.
        TIMED: Unlock now and re-lock after a fixed timeout.
        HOLD: Unlock now and remain unlocked until a lock command is received.
    """

    LOCK = 0x00
    TIMED = 0x01
    HOLD = 0x02


class UserActionType(IntEnum):
    """User action types.

    Attributes:
        SINGLE: User did a single, discrete action, such as pausing a video.
    """

    SINGLE = 0


class ClassifierModelType(IntEnum):
    """Classifier model types.

    Attributes:
        BUILTIN: Model built into the classifier package.
        CUSTOM: Model based on personalized user data.
    """

    BUILTIN = 0
    CUSTOM = 1


class MotionEventType(IntEnum):
    """Types of motion events."""

    TAP = 0


class ClassifierEventType(IntEnum):
    """Types of classifier events."""

    ARM_SYNCED = 0x01
    ARM_UNSYNCED = 0x02
    POSE = 0x03
    UNLOCKED = 0x04
    LOCKED = 0x05
    SYNC_FAILED = 0x06


class Arm(IntEnum):
    """Enumeration identifying a right arm or left arm."""

    RIGHT = 0x01
    LEFT = 0x02
    UNKNOWN = 0xFF


class XDirection(IntEnum):
    """Possible directions for Myo's +x-axis relative to a user's arm."""

    WRIST = 0x01
    ELBOW = 0x02
    UNKNOWN = 0xFF


class SyncResult(IntEnum):
    """Possible outcomes when the user attempts a sync gesture."""

    FAILED_TOO_HARD = 0x01


class FirmwareInfo(NamedTuple):
    """Various parameters that may affect the behaviour of this Myo armband.

    Attributes:
        serial_number: Unique serial number of this Myo.
        unlock_pose: Pose that should be interpreted as the unlock pose.
        active_classifier_type: Whether Myo is currently using a built-in or a custom
            classifier.
        active_classifier_index: Index of the classifier that is currently active.
        has_custom_classifier: Whether Myo contains a valid custom classifier.
        stream_indicating: Set if the Myo uses BLE indicates to stream data, for
            reliable capture.
        sku: SKU value of the device.
    """

    serial_number: bytes
    unlock_pose: Pose
    active_classifier_type: ClassifierModelType
    active_classifier_index: int
    has_custom_classifier: bool
    stream_indicating: bool
    sku: SKU


EmgValue: TypeAlias = Tuple[int, int, int, int, int, int, int, int]


class EMGCallback(Protocol):
    def __call__(self, emg: tuple[EmgValue, EmgValue]) -> None:
        ...


class EMGSmoothCallback(Protocol):
    def __call__(self, emg: EmgValue) -> None:
        ...


class Quaternion(NamedTuple):
    w: float
    x: float
    y: float
    z: float


class IMUCallback(Protocol):
    def __call__(
        self,
        orientation: Quaternion,
        accelerometer: tuple[float, float, float],
        gyroscope: tuple[float, float, float],
    ) -> None:
        ...


class TapCallback(Protocol):
    def __call__(self, direction: int, count: int) -> None:
        ...


class SyncCallback(Protocol):
    def __call__(
        self,
        arm: Arm,
        x_direction: XDirection,
        failed_flag: SyncResult | None = None,
    ) -> None:
        ...


class PoseCallback(Protocol):
    def __call__(self, pose: Pose) -> None:
        ...


class LockCallback(Protocol):
    def __call__(self, locked: bool) -> None:  # noqa: FBT001
        ...



# ruff: noqa: T201, INP001
from __future__ import annotations

import asyncio

from pymyo import Myo
from pymyo.types import EmgMode, EmgValue, ImuMode, SleepMode

# Put your own Myo Bluetooth address here or device UUID if you're on macOS.
# BleakScanner from the bleak library can be used to find it.
MYO_ADDRESS = "D7:91:D9:1C:C3:EB"


async def main() -> None:
    # You can use an asynchronous context manager to manage connection/disconnection
    async with Myo(MYO_ADDRESS) as myo:
        # Access information using awaitable properties
        print("Device name:", await myo.name)
        print("Battery level:", await myo.battery)
        print("Firmware version:", await myo.firmware_version)
        print("Firmware info:", await myo.info)

        await myo.vibrate2((250, 255), (250, 128), (250, 255))

        # Register an event listener
        def on_tap(direction: int, count: int) -> None:
            print(f"Tap: direction: {direction} count: {count}")

        myo.on_tap(on_tap)

        # Register an event listener using a decorator
        @myo.on_emg
        def on_emg(emg: tuple[EmgValue, EmgValue]) -> None:
            print(emg)

        await myo.set_sleep_mode(SleepMode.NEVER_SLEEP)
        await asyncio.sleep(1)
        await myo.set_mode(emg_mode=EmgMode.EMG, imu_mode=ImuMode.EVENTS)

        while True:
            await asyncio.sleep(1)  # Do other stuff


if __name__ == "__main__":
    asyncio.run(main())



"""The Myo client interface module."""
from __future__ import annotations

__all__ = ["Myo"]

import itertools
import struct
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Final,
    Generic,
    TypeVar,
)

if TYPE_CHECKING:
    import sys
    from types import TracebackType

    if sys.version_info < (3, 11):
        from typing_extensions import Self
    else:
        from typing import Self

from bleak import BleakClient

from .types import (
    SKU,
    VIBRATE2_STEPS,
    Arm,
    ClassifierEventType,
    ClassifierMode,
    ClassifierModelType,
    EMGCallback,
    EmgMode,
    EMGSmoothCallback,
    FirmwareInfo,
    FirmwareVersion,
    HardwareRev,
    IMUCallback,
    ImuMode,
    LockCallback,
    Pose,
    PoseCallback,
    Quaternion,
    SleepMode,
    SyncCallback,
    SyncResult,
    TapCallback,
    UnlockType,
    UserActionType,
    VibrationType,
    XDirection,
)

_STANDARD_UUID_FMT: Final = "0000{:04x}-0000-1000-8000-00805f9b34fb"
_MYO_UUID_FMT: Final = "d506{:04x}-a904-deb9-4748-2c7f4a124842"


class _BTChar:
    NAME = _STANDARD_UUID_FMT.format(0x2A00)
    BATTERY = _STANDARD_UUID_FMT.format(0x2A19)
    INFO = _MYO_UUID_FMT.format(0x0101)
    FIRMWARE = _MYO_UUID_FMT.format(0x0201)
    COMMAND = _MYO_UUID_FMT.format(0x0401)
    IMU = _MYO_UUID_FMT.format(0x0402)
    MOTION = _MYO_UUID_FMT.format(0x0502)
    CLASSIFIER = _MYO_UUID_FMT.format(0x0103)
    EMG_SMOOTH = _MYO_UUID_FMT.format(0x0104)
    EMG0 = _MYO_UUID_FMT.format(0x0105)
    EMG1 = _MYO_UUID_FMT.format(0x0205)
    EMG2 = _MYO_UUID_FMT.format(0x0305)
    EMG3 = _MYO_UUID_FMT.format(0x0405)


_C = TypeVar("_C", bound=Callable[..., None])


class Event(Generic[_C]):
    def __init__(self) -> None:
        self._observers: list[_C] = []

    def __call__(self, callback: _C) -> None:
        self._observers.append(callback)

    def notify(self, *args: Any, **kwargs: Any) -> None:
        for observer in self._observers:
            observer(*args, **kwargs)


class Myo:
    """Client used to connect and interact with a Myo armband device.

    All arguments passed to the constructor are forwarded to the underlying BleakClient
    instance.
    (See https://bleak.readthedocs.io/en/latest/api/client.html#bleak.BleakClient)

    Can be used as an asynchronous context manager in order to automatically manage the
    connection and disconnection.

    Attributes:
        on_emg (Event[EMGCallback]): Event for handling EMG data.
        on_emg_smooth (Event[EMGSmoothCallback]): Event for handling smoothed EMG data.
        on_imu (Event[IMUCallback]): Event for handling IMU data.
        on_tap (Event[TapCallback]): Event for handling tap gestures.
        on_sync (Event[SyncCallback]): Event for handling synchronization events.
        on_pose (Event[PoseCallback]): Event for handling pose changes.
        on_lock (Event[LockCallback]): Event for handling lock events.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._device = BleakClient(*args, **kwargs)
        self._emg_mode = EmgMode.NONE
        self._imu_mode = ImuMode.NONE
        self._classifier_mode = ClassifierMode.DISABLED
        self._sleep_mode = SleepMode.NORMAL

        self.on_emg: Event[EMGCallback] = Event()
        self.on_emg_smooth: Event[EMGSmoothCallback] = Event()
        self.on_imu: Event[IMUCallback] = Event()
        self.on_tap: Event[TapCallback] = Event()
        self.on_sync: Event[SyncCallback] = Event()
        self.on_pose: Event[PoseCallback] = Event()
        self.on_lock: Event[LockCallback] = Event()

    async def __aenter__(self) -> Self:
        await self.connect()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        await self.disconnect()

    async def connect(self) -> None:
        """Connect to the device."""
        await self._device.connect()
        await self._device.start_notify(_BTChar.IMU, self._on_imu)
        await self._device.start_notify(_BTChar.MOTION, self._on_motion)
        await self._device.start_notify(_BTChar.CLASSIFIER, self._on_classifier)
        await self._device.start_notify(_BTChar.EMG_SMOOTH, self._on_emg_smooth)
        for c in (_BTChar.EMG0, _BTChar.EMG1, _BTChar.EMG2, _BTChar.EMG3):
            await self._device.start_notify(c, self._on_emg)

    async def disconnect(self) -> None:
        """Disconnect from the device."""
        await self._device.disconnect()

    @property
    def is_connected(self) -> bool:
        """Connection status."""
        return self._device.is_connected

    @property
    async def name(self) -> str:
        """Device name."""
        return (await self._device.read_gatt_char(_BTChar.NAME)).decode()

    @property
    async def battery(self) -> int:
        """Current battery level information in percent."""
        return ord(await self._device.read_gatt_char(_BTChar.BATTERY))

    @property
    async def info(self) -> FirmwareInfo:
        """Various parameters that may affect the behaviour of the device."""
        sn, up, act, aci, hcs, si, sku = struct.unpack(
            "<6sH5B7x",
            await self._device.read_gatt_char(_BTChar.INFO),
        )
        return FirmwareInfo(
            sn,
            Pose(up),
            ClassifierModelType(act),
            aci,
            bool(hcs),
            bool(si),
            SKU(sku),
        )

    @property
    async def firmware_version(self) -> FirmwareVersion:
        """Firmware version information."""
        major, minor, patch, hardware_rev = struct.unpack(
            "<4H",
            await self._device.read_gatt_char(_BTChar.FIRMWARE),
        )
        return FirmwareVersion(major, minor, patch, HardwareRev(hardware_rev))

    @property
    def emg_mode(self) -> EmgMode:
        """Current EMG mode.

        Use `set_mode` to set a new mode.
        """
        return self._emg_mode

    @property
    def imu_mode(self) -> ImuMode:
        """Current IMU mode.

        Use `set_mode` to set a new mode.
        """
        return self._imu_mode

    @property
    def classifier_mode(self) -> ClassifierMode:
        """Current classifier mode.

        Use `set_mode` to set a new mode.
        """
        return self._classifier_mode

    async def set_mode(
        self,
        emg_mode: EmgMode | None = None,
        imu_mode: ImuMode | None = None,
        classifier_mode: ClassifierMode | None = None,
    ) -> None:
        """Set EMG, IMU and classifier modes.

        Args:
            emg_mode (EmgMode | None): The desired EMG mode. If None, the current mode
                will be retained.
            imu_mode (ImuMode | None): The desired IMU mode. If None, the current mode
                will be retained.
            classifier_mode (ClassifierMode | None): The desired classifier mode. If
                None, the current mode will be retained.
        """
        if emg_mode is None:
            emg_mode = self._emg_mode
        if imu_mode is None:
            imu_mode = self._imu_mode
        if classifier_mode is None:
            classifier_mode = self._classifier_mode
        await self._device.write_gatt_char(
            _BTChar.COMMAND,
            struct.pack("<5B", 1, 3, emg_mode, imu_mode, classifier_mode),
            response=False,
        )
        self._emg_mode = emg_mode
        self._imu_mode = imu_mode
        self._classifier_mode = classifier_mode

    @property
    def sleep_mode(self) -> SleepMode:
        """Get the current sleep mode.

        Use `set_sleep_mode` to set a new mode.
        """
        return self._sleep_mode

    async def set_sleep_mode(self, sleep_mode: SleepMode) -> None:
        """Set the sleep mode.

        Args:
            sleep_mode (SleepMode): The desired sleep mode.
        """
        await self._device.write_gatt_char(
            _BTChar.COMMAND,
            struct.pack("<3B", 9, 1, sleep_mode),
            response=False,
        )
        self._sleep_mode = sleep_mode

    async def vibrate(self, vibration_type: VibrationType) -> None:
        """Vibrate according to the desired type of vibration.

        Args:
            vibration_type (VibrationType): The type of vibration.
        """
        await self._device.write_gatt_char(
            _BTChar.COMMAND,
            struct.pack("<3B", 3, 1, vibration_type),
            response=False,
        )

    async def deep_sleep(self) -> None:
        """Put the device into deep sleep.

        Sending this command puts the Myo armband in a mode where all functions are
        shut down. It can remain in this state for months, as it does when initially
        shipped. To reactivate, connect the Myo via USB.

        Note:
            Don't send this command lightly: a user may not know what happened or have
            the ability to recover.
        """
        await self._device.write_gatt_char(_BTChar.COMMAND, b"\x04\x00", response=False)

    async def set_led_colors(
        self,
        logo_rgb: tuple[int, int, int],
        status_rgb: tuple[int, int, int],
    ) -> None:
        """Set the colors for the logo and the status LEDs.

        Note:
            Undocumented in the official API.

        Args:
            logo_rgb (tuple[int, int, int]): RGB values for the logo LED.
            status_rgb (tuple[int, int, int]): RGB values for the status LED bar.

        Example:
            >>> await myo.set_led_colors((255, 0, 0), (0, 255, 0))
        """
        await self._device.write_gatt_char(
            _BTChar.COMMAND,
            struct.pack("<8B", 6, 6, *logo_rgb, *status_rgb),
            response=False,
        )

    async def vibrate2(self, *steps: tuple[int, int]) -> None:
        """Vibrate according to a sequence of customizable steps.

        Args:
            *steps:
                A sequence of vibration steps, each represented by a two-element tuple:
                    - Duration (int): The duration of the vibration in milliseconds.
                    - Strength (int): The strength of the vibration (0: off, 255: max).

        Raises:
            ValueError: If the number of steps exceeds the maximum (VIBRATE2_STEPS).

        Example:
            >>> await myo.vibrate2((100, 128), (200, 192), (300, 255))
        """
        nb_steps = len(steps)
        if nb_steps > VIBRATE2_STEPS:
            msg = f"Expected at most {VIBRATE2_STEPS} vibration steps (got {nb_steps})"
            raise ValueError(msg)
        # Flatten while adding the potentially missing steps
        flat_steps = itertools.chain(*steps, (VIBRATE2_STEPS - nb_steps) * (0, 0))
        await self._device.write_gatt_char(
            _BTChar.COMMAND,
            struct.pack("<2B" + VIBRATE2_STEPS * "HB", 7, 20, *flat_steps),
            response=False,
        )

    async def unlock(self, unlock_type: UnlockType) -> None:
        """Unlock or re-lock the device.

        Args:
            unlock_type (UnlockType): The type of unlock.
        """
        await self._device.write_gatt_char(
            _BTChar.COMMAND,
            struct.pack("<3B", 10, 1, unlock_type),
            response=False,
        )

    async def user_action(
        self,
        action_type: UserActionType = UserActionType.SINGLE,
    ) -> None:
        """Notify user that an action has been recognized / confirmed.

        Args:
            action_type (UserActionType): The type of user action.
        """
        await self._device.write_gatt_char(
            _BTChar.COMMAND,
            struct.pack("<3B", 11, 1, action_type),
            response=False,
        )

    # Notification callbacks
    def _on_emg(self, _: Any, value: bytearray) -> None:
        emg = struct.unpack("<16b", value)
        self.on_emg.notify((emg[:8], emg[8:]))

    def _on_emg_smooth(self, _: Any, value: bytearray) -> None:
        self.on_emg_smooth.notify(struct.unpack("<8Hx", value))

    def _on_imu(self, _: Any, value: bytearray) -> None:
        imu_data = struct.unpack("<10h", value)
        orientation = Quaternion(*[x / 16384 for x in imu_data[:4]])
        accelerometer = tuple(x / 2048 for x in imu_data[4:7])
        gyroscope = tuple(x / 16 for x in imu_data[7:])
        self.on_imu.notify(orientation, accelerometer, gyroscope)

    def _on_motion(self, _: Any, value: bytearray) -> None:
        # The only MotionEventType implemented in the spec is TAP.
        event_type, *tap_data = struct.unpack("<3B", value)
        self.on_tap.notify(*tap_data)

    def _on_classifier(self, _: Any, value: bytearray) -> None:
        event_type, event_data = struct.unpack("<B2s", value)
        if event_type == ClassifierEventType.ARM_SYNCED:
            arm, x_direction = struct.unpack("<2B", event_data)
            self.on_sync.notify(Arm(arm), XDirection(x_direction))
        elif event_type == ClassifierEventType.ARM_UNSYNCED:
            self.on_sync.notify(Arm.UNKNOWN, XDirection.UNKNOWN)
        elif event_type == ClassifierEventType.POSE:
            self.on_pose.notify(Pose(int.from_bytes(event_data, "little")))
        elif event_type == ClassifierEventType.UNLOCKED:
            self.on_lock.notify(False)  # noqa: FBT003
        elif event_type == ClassifierEventType.LOCKED:
            self.on_lock.notify(True)  # noqa: FBT003
        elif event_type == ClassifierEventType.SYNC_FAILED:
            # The only SyncResult implemented in the spec is FAILED_TOO_HARD.
            self.on_sync.notify(
                Arm.UNKNOWN, XDirection.UNKNOWN, SyncResult.FAILED_TOO_HARD
            )
        else:
            # Should never happen, unless spec changes.
            msg = f"Invalid ClassifierEventType: {event_type}"
            raise RuntimeError(msg)
